import logging
import os
import re
from typing import Any, List, Optional, Set, Type, Union

from mloda_core.abstract_plugins.abstract_feature_group import AbstractFeatureGroup
from mloda_core.abstract_plugins.components.feature import Feature
from mloda_core.abstract_plugins.components.feature_set import FeatureSet
from mloda_core.abstract_plugins.components.input_data.base_input_data import BaseInputData
from mloda_core.abstract_plugins.components.input_data.creator.data_creator import DataCreator
from mloda_core.abstract_plugins.compute_frame_work import ComputeFrameWork
from mloda_core.api.request import mlodaAPI
from mloda_plugins.compute_framework.base_implementations.pandas.dataframe import PandasDataframe
from mloda_plugins.feature_group.experimental.llm.llm_api.gemini import GeminiRequestLoop
from mloda_plugins.feature_group.experimental.llm.llm_file_selector import LLMFileSelector
from mloda_plugins.feature_group.experimental.llm.tools.available.adjust_file_tool import AdjustFileTool
from mloda_plugins.feature_group.experimental.llm.tools.available.git_diff import GitDiffTool
from mloda_plugins.feature_group.experimental.llm.tools.available.git_diff_cached import GitDiffCachedTool
from mloda_plugins.feature_group.experimental.llm.tools.available.read_file_tool import ReadFileTool
from mloda_plugins.feature_group.experimental.llm.tools.available.replace_file_tool import ReplaceFileTool
from mloda_plugins.feature_group.experimental.llm.tools.available.run_tox import RunToxTool
from mloda_plugins.feature_group.experimental.llm.tools.base_tool import BaseTool
from mloda_plugins.feature_group.experimental.llm.tools.tool_collection import ToolCollection
from mloda_plugins.feature_group.experimental.llm.tools.tool_data_classes import PytestResult
from mloda_plugins.feature_group.input_data.read_context_files import ConcatenatedFileContent
from mloda_plugins.feature_group.experimental.default_options_key import DefaultOptionKeys

logger = logging.getLogger(__name__)


class RunRefactorDiffCached:
    def __init__(self) -> None:
        self.compute_frameworks: Set[Type[ComputeFrameWork]] = {PandasDataframe}

    def run(self) -> None:
        # check tests are passing
        self.run_tox_feature_group()

        # get related file to git diff
        get_diff_cached = self.get_tool_output_by_feature_group_(DiffCachedFeatureGroup)
        diff_cache_relevant_files = self.get_relevant_files_for_refactoring(get_diff_cached)
        related_files_to_git_diff = self.get_related_files_to_diff_cached(diff_cache_relevant_files)

        # identify code smells in the actual code
        split_files = self.split_related_files(related_files_to_git_diff)
        single_code_smell = self.identify_one_code_smell(split_files, get_diff_cached)

        # check if tests are passing
        previous = ""
        actual_git_diff = self.get_tool_output_by_feature_group_(DiffFeatureGroup)

        for i in range(5):
            previous = self.fix_code_smell(split_files, single_code_smell, previous, actual_git_diff)
            if "AnalysisComplete" in previous:
                break

    def fix_code_smell(self, files: List[str], code_smell: str, previous: str, current_git_diff: str) -> str:
        if previous:
            _previous = f"""**Previous Refactoring Steps:**
                            {previous}
            """
        else:
            _previous = """ **Previous Refactoring Steps:**
                              No previous refactoring steps have been taken.
                        """

        prompt = f""" 

                **Objective:**

                Your objective is to automatically refactor code to eliminate a specific code smell while ensuring no existing functionality is broken.  Crucially, your refactoring must *not* involve adding or modifying project dependencies.

                **Task Description:**

                You are an automated code refactoring agent. You will be given information about a codebase and a specific code smell to address. Your task is to iteratively refactor the code using provided tools until the code smell is resolved and all tests pass.

                For each iteration, you will receive:

                1.  **Code Smell Description:**  A textual description of the code smell, including its location (file and potentially line numbers or code snippet).
                2.  **Previous Refactoring Steps:** A summary of refactoring actions already taken (if any). This is for informational context.
                3.  **Current Code Difference (Git Diff):** A `git diff` showing the changes made to the initial codebase so far. This represents the current codebase state.
                4.  **Initial Code Files:** Access to the original codebase files (for reference, though you will be working with the current state as reflected in the diff).

                **Your Actions in each iteration:**

                **Your Actions in each iteration:**

                1.  **Analyze the Code Smell:**  Understand the description and location of the code smell.
                2.  **Plan Refactoring Action:** Based on the code smell and the current code state, determine a *specific refactoring action* to take.  This action must be achievable using the provided tools (`AdjustFileTool`, `ReplaceFileTool`, `CreateFileTool`, `ReadFileTool`). Prioritize actions that directly address the described code smell. Consider if reading file content (`ReadFileTool`) is necessary to inform your refactoring plan.  *Remember: No dependency changes allowed.*
                3.  **Select Tool and Target File(s):** Choose the appropriate tool (`AdjustFileTool`, `ReplaceFileTool`, `CreateFileTool`, or `ReadFileTool`) and identify the file(s) you need to interact with. For `CreateFileTool` and `ReplaceFileTool`, ensure you have the new file content ready. For `AdjustFileTool`, determine the specific changes needed within the file. For `ReadFileTool`, specify the file to read.
                4.  **Execute Tool:**  Use the chosen tool to modify or read the file(s) according to your planned refactoring action.  Provide the necessary parameters for the tool (e.g., file path, content changes for modification tools, file path for reading).
                5.  **Run Tests:** After each code modification (using `AdjustFileTool`, `ReplaceFileTool`, or `CreateFileTool`), immediately use the `RunToxTool` to execute the test suite.  *Note: `ReadFileTool` does not modify code, so no test run is needed immediately after using it, unless it's part of a larger refactoring plan that involves code modification in the same iteration.*
                6.  **Evaluate Test Results:** Check if all tests passed (after using modification tools).
                    *   **If tests pass:** Proceed to the next iteration. Analyze if the code smell is resolved. If the code smell is considered resolved, output "AnalysisComplete". If not resolved, continue iterating from step 1, considering the current code state and the remaining code smell.
                    *   **If tests fail:** Your refactoring action has broken functionality. **Instead of immediately reverting, attempt to diagnose and fix the error.**
                        *   **6.a. Analyze Test Output:** Examine the output from `RunToxTool`. Look for error messages, stack traces, and failed test names. Try to identify the *likely cause* of the test failure based on this output.  *(Initially, this can be simple keyword matching or pattern recognition in error messages.  For example, look for "NameError", "TypeError", "AssertionError", file paths, or line numbers mentioned in the errors.)*
                        *   **6.b. Plan Fix Action (Attempt 1):** Based on your analysis of the test output, formulate a *targeted fix action*. This should be a small, focused change aimed at addressing the likely cause of the failure.  Use `AdjustFileTool` to make this fix.  *(For example, if you see a "NameError" mentioning a variable name, you might try to rename the variable back to its original name or correct a typo in its usage.)*
                        *   **6.c. Execute Fix and Re-run Tests:** Execute the planned fix using `AdjustFileTool` and then immediately run `RunToxTool` again to see if the fix resolved the issue.
                        *   **6.d. Evaluate Fix Attempt:**
                            *   **If tests pass after the fix:**  The error is resolved! Proceed to the next iteration, analyzing if the code smell is now resolved and continuing the refactoring process.
                            *   **If tests still fail after the fix attempt:** The initial fix was not successful. **Now, revert to the code state *before* the *original refactoring action that caused the tests to fail* (not just the fix attempt). Output "TestFailedAfterFixAttempt".** In more advanced scenarios, you could try a different fix, or log the failure for human review. For now, reverting after one fix attempt is a reasonable level of complexity.
                                **Tools Available:**

                *   `AdjustFileTool`:  Modify specific parts of an existing file.  Requires specifying the file path and the changes to be made (e.g., line number, content to replace, content to insert).
                *   `CreateFileTool`: Create a new file with the specified content. Requires specifying the file path and the new file content.
                *   `ReplaceFileTool`: Replace the entire content of an existing file. Requires specifying the file path and the new file content. *NOTE: DO NOT RUN THIS TOOL IF THE FILE DOES NOT EXISTS.*
                *   `RunToxTool`: Execute the test suite.  No parameters needed. Returns pass/fail status.
                *   `ReadFileTool`: Read the content of a file. Requires specifying the exact file path.

                **Constraints:**

                *   **No Dependency Changes:** You *cannot* modify any dependency-related files or add new dependencies.
                *   **Functionality Must Be Preserved:**  All existing tests must pass after each successful refactoring step when code is modified.

                **Input Data Format (for each iteration):**
                Code Smell:
                {code_smell}

                {_previous}

                Current Difference to the Codebase:
                {current_git_diff}

                **Initial Code Files**:
                """  # nosec

        tool_collection = ToolCollection()
        tool_collection.add_tool(AdjustFileTool.get_class_name())
        tool_collection.add_tool(RunToxTool.get_class_name())
        tool_collection.add_tool(ReplaceFileTool.get_class_name())
        tool_collection.add_tool(ReadFileTool.get_class_name())

        feature = Feature(
            name=GeminiRequestLoop.get_class_name(),
            options={
                "model": "gemini-2.0-flash-exp",  # Choose your desired model
                "prompt": prompt,
                DefaultOptionKeys.mloda_source_feature: frozenset([ConcatenatedFileContent.get_class_name()]),
                "file_paths": frozenset(files),
                "project_meta_data": True,
                "tools": tool_collection,
            },
        )

        results = mlodaAPI.run_all(
            [feature],
            compute_frameworks=self.compute_frameworks,
        )
        res = results[0][GeminiRequestLoop.get_class_name()].values[0]
        res = previous + res

        if isinstance(res, str):
            print(res)
            return res
        raise ValueError("Wrong type of result")

    def split_related_files(self, related_files_to_git_diff: str) -> List[str]:
        files = related_files_to_git_diff.split(",")
        new_files = []
        for file in files:
            new_files.append(file.strip("\n"))
        return new_files

    def identify_one_code_smell(self, files: List[str], git_diff_cached: str) -> str:
        """
        - Potential performance bottlenecks (e.g., inefficient algorithms, unnecessary loops)
            - Dead code (unused variables, functions, or code paths)
            - General code smells (e.g., long parameter lists, excessive class coupling)
            - Functions with high cyclomatic complexity

        """

        prompt = f""" 
            You are an experienced software engineer specializing in code refactoring and quality analysis. Your goal is to identify a *specific, actionable* refactoring target introduced or exacerbated by the code changes described in the following `git diff --cached`.

            Given the following `git diff --cached` output:
            
            {git_diff_cached}

            and the following list of potential code smells:

            - Duplicated code blocks
            - Readability issues (unclear variable names)
            
            DO NOT INCLUDE code smells related to options.

            1.  **Analyze the `git diff --cached` output and the code it modifies.**
            2.  **Identify *one specific* code smell that is newly introduced or significantly worsened by the changes in the `git diff`.**  Focus on problems that are clearly fixable and would provide a tangible benefit to the codebase.
            3.  **Create a concise summary (no more than 100 words) that includes:**
                *   A clear description of the identified code smell.
                *   The specific location(s) in the code where the code smell occurs (e.g., file name, function name, line numbers).
                *   A brief explanation of *why* this is a code smell and what negative impact it has.
                *   A *brief* suggestion for how to refactor the code to address the smell.

            If no new or significantly worsened code smells are apparent in the `git diff`, respond with "No actionable refactoring target identified."

        """
        print()
        feature = Feature(
            name=GeminiRequestLoop.get_class_name(),
            options={
                "model": "gemini-2.0-flash-exp",  # Choose your desired model
                "prompt": prompt,
                DefaultOptionKeys.mloda_source_feature: frozenset([ConcatenatedFileContent.get_class_name()]),
                "file_paths": frozenset(files),
                "project_meta_data": True,
            },
        )

        results = mlodaAPI.run_all(
            [feature],
            compute_frameworks=self.compute_frameworks,
        )
        res = results[0][GeminiRequestLoop.get_class_name()].values[0]

        if isinstance(res, str):
            print(res)
            return res
        raise ValueError("Wrong type of result")

    def get_tool_output_by_feature_group_(self, tool_feature_group: Type[AbstractFeatureGroup]) -> str:
        _feature_name = tool_feature_group.get_class_name()
        results = mlodaAPI.run_all(
            [_feature_name],
            compute_frameworks=self.compute_frameworks,
        )
        res = results[0][_feature_name].values[0]

        if isinstance(res, str):
            return res
        raise ValueError("Wrong type of result")

    def run_tox_feature_group(self) -> None:
        print("Start tox")
        _feature_name = ToxFeatureGroup.get_class_name()
        mlodaAPI.run_all(
            [_feature_name],
            compute_frameworks=self.compute_frameworks,
        )
        print("Tox tests passed")

    def get_related_files_to_diff_cached(self, relevant_files: str) -> str:
        prompt = f"""
                You are an experienced software engineer specializing in code refactoring and quality analysis. Your goal is to identify the most relevant code files for addressing specific refactoring concerns.

                Given the following code files, which are 10 most relevant files to answer refactoring questions such as:
                - Are there any duplicated code blocks?
                - Can the code be made more readable?
                - Are there potential performance bottlenecks?
                - Is there any dead code?
                - Are there any code smells?
                - Are there functions with high cyclomatic complexity?

               to following given files: {relevant_files}. 
               List the whole path of the file, separated by commas without any other chars."""

        target_folder = [
            os.getcwd() + "/mloda_plugins",
            os.getcwd() + "/mloda_core",
            os.getcwd() + "/tests/test_plugins",
        ]

        feature: str | Feature = Feature(
            name=LLMFileSelector.get_class_name(),
            options={
                "prompt": prompt,
                "target_folder": frozenset(target_folder),
                "disallowed_files": frozenset(
                    [
                        "__init__.py",
                    ]
                ),
                "file_type": "py",
                "project_meta_data": True,
            },
        )

        results = mlodaAPI.run_all(
            [feature],
            compute_frameworks=self.compute_frameworks,
        )

        res = results[0][LLMFileSelector.get_class_name()].values[0]
        if isinstance(res, str):
            print(res)
            return res
        raise ValueError("Wrong type of result")

    def get_relevant_files_for_refactoring(self, git_diff_cached: str) -> str:
        """
        Identifies relevant files from a git diff string for refactoring purposes.

        Args:
            git_diff_cached: A string containing the output of git diff --cached.

        Returns:
            A comma-separated string of file paths deemed relevant for refactoring.
        """

        relevant_files = set()
        file_paths = re.findall(r"diff --git a/(.*) b/\1", git_diff_cached)
        for file_path in file_paths:
            # Heuristic 1: Check for test file modifications.  Refactoring often
            # involves changes to tests.
            if "test" in file_path:
                relevant_files.add(file_path)
                continue  # Skip further checks for test files

            # Heuristic 2: Large changes suggest significant code modification,
            # potentially requiring refactoring.
            added_lines = git_diff_cached.count("+")
            removed_lines = git_diff_cached.count("-")
            total_changes = added_lines + removed_lines
            if total_changes > 10:  # Adjust threshold as needed
                relevant_files.add(file_path)
                continue

            # Heuristic 3: Look for class or function definitions/modifications.
            # These often indicate structural changes that benefit from refactoring.
            if re.search(r"^\+(class|def)\s+\w+\(", git_diff_cached, re.MULTILINE):  # Added class/function
                relevant_files.add(file_path)
                continue

            if re.search(r"^\-(class|def)\s+\w+\(", git_diff_cached, re.MULTILINE):  # Removed class/function
                relevant_files.add(file_path)
                continue

            # Heuristic 4: Check for import changes.  These may affect dependecies.
            if re.search(r"^\+import", git_diff_cached, re.MULTILINE):
                relevant_files.add(file_path)
                continue

        return ",".join(sorted(list(relevant_files)))


class RunToolFeatureGroup(AbstractFeatureGroup):
    _tool: Type[BaseTool] | None = None

    @classmethod
    def input_data(cls) -> Optional[BaseInputData]:
        return DataCreator({cls.get_class_name()})

    @classmethod
    def compute_framework_rule(cls) -> Union[bool, Set[Type[ComputeFrameWork]]]:
        return {PandasDataframe}

    @classmethod
    def calculate_feature(cls, data: Any, features: FeatureSet) -> Any:
        result_stdout = cls.get_tool().execute()
        result_string = cls.get_tool().create_result_string(result_stdout)
        return {cls.get_class_name(): [result_string]}

    @classmethod
    def get_tool(cls) -> Type[BaseTool]:
        if cls._tool is None:
            raise NotImplementedError("Tool not implemented.")
        return cls._tool


class DiffCachedFeatureGroup(RunToolFeatureGroup):
    _tool = GitDiffCachedTool

    @classmethod
    def validate_output_features(cls, data: Any, features: FeatureSet) -> Optional[bool]:
        if data[cls.get_class_name()].values[0] == cls.get_tool().create_result_string(""):
            raise ValueError("No staged changes found in the repository.")
        return True


class DiffFeatureGroup(RunToolFeatureGroup):
    _tool = GitDiffTool


class ToxFeatureGroup(AbstractFeatureGroup):
    @classmethod
    def compute_framework_rule(cls) -> Union[bool, Set[Type[ComputeFrameWork]]]:
        return {PandasDataframe}

    @classmethod
    def calculate_feature(cls, data: Any, features: FeatureSet) -> Any:
        result = RunToxTool.execute()
        result_string = RunToxTool.create_result_string(result)

        if not isinstance(result, PytestResult):
            raise ValueError("Wrong type of result")

        if result.return_code == 0:
            return {cls.get_class_name(): [result_string]}
        raise ValueError(f"Tox tests failed: {result_string}")



from typing import Any

import pyarrow as pa

from mloda_core.abstract_plugins.components.index.index import Index
from mloda_core.abstract_plugins.components.link import JoinType
from mloda_core.abstract_plugins.components.merge.base_merge_engine import BaseMergeEngine


class PyArrowMergeEngine(BaseMergeEngine):
    def merge_inner(self, left_data: Any, right_data: Any, left_index: Index, right_index: Index) -> Any:
        return self.join_logic("inner", left_data, right_data, left_index, right_index, JoinType.INNER)

    def merge_left(self, left_data: Any, right_data: Any, left_index: Index, right_index: Index) -> Any:
        return self.join_logic("left outer", left_data, right_data, left_index, right_index, JoinType.LEFT)

    def merge_right(self, left_data: Any, right_data: Any, left_index: Index, right_index: Index) -> Any:
        return self.join_logic("right outer", left_data, right_data, left_index, right_index, JoinType.RIGHT)

    def merge_full_outer(self, left_data: Any, right_data: Any, left_index: Index, right_index: Index) -> Any:
        return self.join_logic("full outer", left_data, right_data, left_index, right_index, JoinType.OUTER)

    def merge_append(self, left_data: Any, right_data: Any, left_index: Index, right_index: Index) -> Any:
        # Ensure the schemas of both tables match before appending
        if left_data.schema != right_data.schema:
            raise ValueError("Schemas of the tables do not match for append operation.")
        return pa.concat_tables([left_data, right_data])

    def merge_union(self, left_data: Any, right_data: Any, left_index: Index, right_index: Index) -> Any:
        """
        https://github.com/apache/arrow/issues/30950 Currently, not existing in base pyarrow.
        If needed, one could add it.
        """
        raise ValueError(f"JoinType union are not yet implemented {self.__class__.__name__}")

    def join_logic(
        self, join_type: str, left_data: Any, right_data: Any, left_index: Index, right_index: Index, jointype: JoinType
    ) -> Any:
        if left_index.is_multi_index() or right_index.is_multi_index():
            raise ValueError(f"MultiIndex is not yet implemented {self.__class__.__name__}")

        if left_index.index[0] != right_index.index[0]:
            # PyArrow drops the index column in all cases.
            # Thus, we create a copy of the index column and append it to the right_data to avoid this in case of different index columns.
            _right_index = "mloda_right_index"
            if _right_index in right_data.column_names:
                raise ValueError(f"Column name {_right_index} already exists in right_data.")

            right_data = right_data.append_column(_right_index, right_data[right_index.index[0]])
        else:
            _right_index = right_index.index[0]

        left_data = left_data.join(
            right_data,
            keys=[left_index.index[0]],
            right_keys=[_right_index],
            join_type=join_type,
        )

        return left_data



from typing import Any, Tuple

from pyarrow import feather as pyarrow_feather

from mloda_core.abstract_plugins.components.feature_set import FeatureSet
from mloda_plugins.feature_group.input_data.read_file import ReadFile


class FeatherReader(ReadFile):
    @classmethod
    def suffix(cls) -> Tuple[str, ...]:
        return (
            ".feather",
            ".feather",
        )

    @classmethod
    def load_data(cls, data_access: Any, features: FeatureSet) -> Any:
        return pyarrow_feather.read_table(source=data_access, columns=list(features.get_all_names()))



from copy import copy
from dataclasses import asdict
import json
import os
import time
from typing import Any, Dict, List, Optional, Tuple, Union


from mloda_core.abstract_plugins.components.feature_set import FeatureSet
from mloda_plugins.feature_group.experimental.llm.llm_api.llm_base_request import LLMBaseApi
from mloda_plugins.feature_group.experimental.llm.llm_api.request_loop import RequestLoop
from mloda_plugins.feature_group.experimental.llm.tools.tool_collection import ToolCollection
from mloda_plugins.feature_group.experimental.llm.tools.tool_data_classes import ToolFunctionDeclaration

try:
    from openai import OpenAI
except ImportError:
    OpenAI = None

try:
    import pandas as pd
except ImportError:
    pd = None


import logging

logger = logging.getLogger(__name__)


def python_type_to_openapi_type(python_type: str) -> str:
    """Converts Python type strings to OpenAPI/JSON Schema type strings."""
    type_mapping = {
        "float": "number",
        "int": "integer",
        "str": "string",
        "bool": "boolean",
        "number": "number",
    }
    return type_mapping.get(python_type, "string")  # Default to string if not found


def parse_tool_function_for_openai(function_declaration: ToolFunctionDeclaration) -> Dict[str, Any]:
    """Parses a ToolFunctionDeclaration into a dict formatted for OpenAI function calling.

    The output will have the following structure:
    {
        "type": "function",
        "function": {
            "name": <function name>,
            "description": <function description>,
            "parameters": {
                "type": "object",
                "properties": {
                    <param_name>: {
                        "type": <openai json schema type>,
                        "description": <param description>
                    },
                    ...
                },
                "required": [<list of required param names>],
                "additionalProperties": False
            },
            "strict": True
        }
    }
    """
    if not isinstance(function_declaration, ToolFunctionDeclaration):
        raise ValueError(f"Tool function {function_declaration} does not return a ToolFunctionDeclaration.")

    # Build the parameters schema.
    parameters_schema = {
        "type": "object",
        "properties": {},
        "required": function_declaration.required,
        "additionalProperties": False,
    }

    for param in function_declaration.parameters:
        param_dict = asdict(param)
        parameters_schema["properties"][param_dict["name"]] = {  # type: ignore
            "type": python_type_to_openapi_type(param_dict["type"]),
            "description": param_dict["description"],
        }

    # Construct the final tool structure.
    tool_dict = {
        "type": "function",
        "function": {
            "name": function_declaration.name,
            "description": function_declaration.description,
            "parameters": parameters_schema,
            "strict": True,
        },
    }

    return tool_dict


class OpenAIAPI(LLMBaseApi):
    @classmethod
    def request(
        cls,
        model: str,
        prompt: Union[str, List[Dict[str, str]]],
        model_parameters: Dict[str, Any],
        tools: ToolCollection | None,
    ) -> Any:
        if isinstance(prompt, str):
            raise ValueError("OpenAI requires a list of messages, not a single prompt.")

        try:
            openai_client = cls._setup_model_if_needed(model)
            if openai_client is not None:
                tools = cls.parse_tools(tools)  # type: ignore
                result = cls.generate_response(openai_client, model, prompt, tools)
                return result
        except Exception as e:
            logger.error(f"Error during OpenAI request: {e}")
            raise
        raise ValueError("OpenAI model is not set.")

    @classmethod
    def _setup_model_if_needed(cls, model: str) -> Any:
        api_key = os.environ.get("GEMINI_API_KEY")
        if not api_key:
            raise ValueError("OPENAI_API_KEY environment variable is not set.")

        client = OpenAI(api_key=api_key, base_url="https://generativelanguage.googleapis.com/v1beta/openai/")
        return client

    @classmethod
    def parse_tools(cls, tools: ToolCollection | None) -> List[Dict[str, Any]]:
        """Parses all tools in the ToolCollection for OpenAI."""
        parsed_tools: List[Dict[str, Any]] = []
        if tools is None:
            return parsed_tools
        for _, tool in tools.get_all_tools().items():
            parsed_tools.append(parse_tool_function_for_openai(tool))
        return parsed_tools

    @classmethod
    def handle_response(
        cls, response: Any, features: FeatureSet, tools: ToolCollection | None
    ) -> Tuple[str, Optional[str]]:
        response_text = ""
        tool_result = None

        if hasattr(response, "choices") and len(response.choices) > 0:  # OpenAI response structure
            message = response.choices[0].message
            if message.content:
                response_text = message.content

            if message.tool_calls:  # Check for function call
                if tools is None:
                    raise ValueError("Tools are not set.")

                new_tools = []
                for e in message.tool_calls:
                    tool_dict = {"name": e.function.name, "args": json.loads(e.function.arguments)}
                    new_tools.append(tool_dict)

                tool_result = cls._execute_tools(new_tools, features, tools)
                return "", tool_result

        elif hasattr(response, "text"):  # Fallback - might not be needed for chat models but for robustness
            response_text = response.text
        else:
            logger.warning(f"Response has no text or choices attribute: {response}")
            return str(response), None

        return response_text, None

    def generate_response(
        client: Any,
        model: str,
        messages: List[Dict[str, str]],
        tools: Any,
        max_retries: int = 5,
        initial_retry_delay: int = 10,
        max_retry_delay: int = 60,
    ) -> Any:
        """
        Generates content from OpenAI with retry logic for rate limits.
        """
        # Override defaults with environment variables if present
        max_retries = int(os.environ.get("OPENAI_MAX_RETRIES", str(max_retries)))
        initial_retry_delay = int(os.environ.get("OPENAI_INITIAL_RETRY_DELAY", str(initial_retry_delay)))
        max_retry_delay = int(os.environ.get("OPENAI_MAX_RETRY_DELAY", str(max_retry_delay)))

        retry_attempt = 0
        while retry_attempt <= max_retries:
            try:
                result = client.chat.completions.create(model=model, messages=messages, tools=tools)
                return result
            except Exception as e:
                # Check for an OpenAI rate limit error; adjust the error check as needed
                is_rate_limit_error = False
                if e.code == 429:  # type: ignore
                    is_rate_limit_error = True

                if is_rate_limit_error:
                    retry_attempt += 1
                    if retry_attempt > max_retries:
                        print(f"Maximum retries ({max_retries}) reached for OPENAI. Raising exception.")
                        raise
                    delay = min(initial_retry_delay * (2 ** (retry_attempt - 1)), max_retry_delay)
                    print(
                        f"Rate limit hit for OPENAI. Retrying in {delay:.2f} seconds (Attempt {retry_attempt}/{max_retries})."
                    )
                    time.sleep(delay)
                else:
                    print(f"An unexpected error occurred during OPENAI request: {e}")
                    raise

        raise Exception(f"Maximum retries ({max_retries}) reached for OPENAI without a successful response.")


class OpenAIRequestLoop(RequestLoop):
    @classmethod
    def api(cls) -> Any:
        return OpenAIAPI

    @classmethod
    def calculate_feature(cls, data: Any, features: FeatureSet) -> Any:
        model, initial_prompt, model_parameters, tools = cls.read_properties(data, features)

        current_prompt = copy(initial_prompt)

        all_results = []
        messages: List[Dict[str, str]] = []  # Initialize messages for OpenAI

        all_tool_result = ""
        while True:
            print("\n############################################\n")
            api_instance = cls.api()

            if not messages:
                messages = [{"role": "user", "content": current_prompt}]

            response = api_instance.request(model, messages, model_parameters, tools)

            response_text, tool_result = cls.api().handle_response(response, features, tools)

            if response_text:
                response_text = response_text.strip()
            all_results.append(response_text)

            if not tool_result:
                break
            else:
                print(tool_result)

                messages = cls.loop_prompt(messages, tool_result, all_tool_result)

            all_tool_result += tool_result
        return pd.DataFrame({cls.get_class_name(): ["\n".join(all_results)]})

    @classmethod
    def loop_prompt(
        cls, prompt_or_messages: Any, tool_result: str, all_tool_result: str
    ) -> Any:  # Accepting either prompt string or messages list
        messages: List[Dict[str, str]] = prompt_or_messages  # Type hint for clarity
        messages.append(
            {"role": "assistant", "content": f"Tool Result: {tool_result}"}
        )  # Add tool result as assistant message
        messages.append(
            {
                "role": "user",
                "content": """
            **Instructions:**

            You are an expert reasoning agent. Given the information above (the original instructions, prior steps, and the most recent step result), carefully analyze the situation and determine the next action to take.

            *   If the goal is complete, respond with the `Final Answer: ` followed by the final answer.
            *   If another tool is needed, determine the correct tool to use, and what input it needs.
            """,
            }
        )  # Add user instruction for next turn, incorporating previous instructions
        return messages  # Return the updated messages list



from typing import Any, Tuple

from pyarrow import json as pyarrow_json

from mloda_core.abstract_plugins.components.feature_set import FeatureSet
from mloda_plugins.feature_group.input_data.read_file import ReadFile


class JsonReader(ReadFile):
    @classmethod
    def suffix(cls) -> Tuple[str, ...]:
        return (
            ".json",
            ".JSON",
        )

    @classmethod
    def load_data(cls, data_access: Any, features: FeatureSet) -> Any:
        result = pyarrow_json.read_json(
            data_access,
            parse_options=pyarrow_json.ParseOptions(
                explicit_schema=None,
                unexpected_field_behavior="error",
            ),
        )
        return result.select(list(features.get_all_names()))

    @classmethod
    def get_column_names(cls, file_name: str) -> Any:
        # Read only the first batch of rows to infer the schema
        read_options = pyarrow_json.ReadOptions(block_size=65536)  # Reads a small sample
        table = pyarrow_json.read_json(file_name, read_options=read_options)
        return table.schema.names



from typing import Any, Optional
from mloda_core.abstract_plugins.abstract_feature_group import AbstractFeatureGroup
from mloda_core.abstract_plugins.components.feature_set import FeatureSet
from mloda_core.abstract_plugins.components.input_data.api.api_input_data import ApiInputData
from mloda_core.abstract_plugins.components.input_data.base_input_data import BaseInputData


class ApiInputDataFeature(AbstractFeatureGroup):
    @classmethod
    def input_data(cls) -> Optional[BaseInputData]:
        return ApiInputData()

    @classmethod
    def calculate_feature(cls, data: Any, features: FeatureSet) -> Any:
        return data



from typing import Any, Set, Type
from mloda_core.abstract_plugins.components.merge.base_merge_engine import BaseMergeEngine
from mloda_plugins.compute_framework.base_implementations.pyarrow.pyarrow_merge_engine import PyArrowMergeEngine
import pyarrow as pa

from mloda_core.abstract_plugins.components.cfw_transformer import ComputeFrameworkTransformMap
from mloda_core.abstract_plugins.components.feature_name import FeatureName
from mloda_core.abstract_plugins.compute_frame_work import ComputeFrameWork


try:
    import pandas as pd
except ImportError:
    pd = None


class PyarrowTable(ComputeFrameWork):
    @staticmethod
    def expected_data_framework() -> Any:
        return pa.Table

    def merge_engine(self) -> Type[BaseMergeEngine]:
        return PyArrowMergeEngine

    def transform(
        self,
        data: Any,
        feature_names: Set[str],
        transform_map: ComputeFrameworkTransformMap = ComputeFrameworkTransformMap(),
    ) -> Any:
        transformed_data = self.transform_refactored(data, transform_map)
        if transformed_data is not None:
            return transformed_data

        if isinstance(data, dict):
            """Initial data: Transform dict to table"""
            return pa.table(data)

        if isinstance(data, pa.ChunkedArray) or isinstance(data, pa.Array):
            """Added data: Add column to table"""
            if len(feature_names) == 1:
                return self.data.append_column(next(iter(feature_names)), data)
            raise ValueError(f"Only one feature can be added at a time: {feature_names}")

        raise ValueError(f"Data {type(data)} is not supported by {self.__class__.__name__}")

    def select_data_by_column_names(self, data: Any, selected_feature_names: Set[FeatureName]) -> Any:
        return data.select([f.name for f in selected_feature_names])

    def set_column_names(self) -> None:
        self.column_names = set(self.data.schema.names)



import subprocess  # nosec
import sys
from typing import Any, Set, Type, Union

from mloda_core.abstract_plugins.abstract_feature_group import AbstractFeatureGroup

from mloda_core.abstract_plugins.components.feature_set import FeatureSet
from mloda_core.abstract_plugins.compute_frame_work import ComputeFrameWork
from mloda_plugins.compute_framework.base_implementations.pandas.dataframe import PandasDataframe


class InstalledPackagesFeatureGroup(AbstractFeatureGroup):
    @classmethod
    def calculate_feature(cls, data: Any, features: FeatureSet) -> Any:
        try:
            result = subprocess.run([sys.executable, "-m", "pip", "freeze"], capture_output=True, text=True, check=True)  # nosec
            packages = result.stdout
            return {cls.get_class_name(): [packages]}
        except subprocess.CalledProcessError as e:
            error_message = f"Command '{e.cmd}' failed with return code {e.returncode}. Error output: {e.stderr}"
            return {"error": error_message}

    @classmethod
    def compute_framework_rule(cls) -> Union[bool, Set[Type[ComputeFrameWork]]]:
        return {PandasDataframe}



from __future__ import annotations
from typing import Generator, List, Optional, Set, Union
from uuid import UUID
from mloda_core.abstract_plugins.components.feature import Feature
from mloda_core.abstract_plugins.components.options import Options
from mloda_plugins.feature_group.experimental.default_options_key import DefaultOptionKeys


class Features:
    """A class to create the features collection and do basic validation.

    Parent uuid is internal logic for now. Don t give this parameter.
    """

    def __init__(
        self,
        features: List[Union[Feature, str]],
        child_options: Optional[Options] = None,
        child_uuid: Optional[UUID] = None,
    ) -> None:
        self.collection: List[Feature] = []
        self.child_uuid: Optional[UUID] = child_uuid

        self.parent_uuids: set[UUID] = set()

        if child_options is None:
            child_options = Options({})

        self.check_for_duplicate_string_features(features)
        self.build_feature_collection(features, child_options, child_uuid)

    def build_feature_collection(
        self, features: List[Union[Feature, str]], child_options: Options, child_uuid: Optional[UUID] = None
    ) -> None:
        for feature in features:
            if child_options.data == {}:
                child_options = Options({})

            feature = Feature(name=feature, options=child_options) if isinstance(feature, str) else feature

            if child_uuid:
                self.parent_uuids.add(feature.uuid)
                self.child_uuid = child_uuid
                feature.child_options = child_options
                self.merge_options(feature.options, child_options)

            self.check_duplicate_feature(feature)
            self.collection.append(feature)

    def merge_options(self, feature_options: Options, child_options: Options) -> None:
        """Merge child_options into feature_options without overwriting existing keys."""

        for key_child, value_child in child_options.data.items():
            for key_parent, value_parent in feature_options.data.items():
                if key_child == key_parent:
                    if key_parent == DefaultOptionKeys.mloda_source_feature:
                        continue
                    if value_child != value_parent:
                        raise ValueError(f"Duplicate keys found in options: {key_child}")

        # Merge the child_options data into feature_options data
        feature_options.update_considering_mloda_source(child_options)

    def check_duplicate_feature(self, feature: Feature) -> None:
        if feature in self.collection:
            raise ValueError(f"Duplicate feature setup: {feature.name}")

    def __iter__(self) -> Generator[Feature, None, None]:
        yield from self.collection

    def check_for_duplicate_string_features(self, features: List[Union[Feature, str]]) -> None:
        check_set: Set[str] = set()
        for feature in features:
            if isinstance(feature, str):
                if feature in check_set:
                    raise ValueError(f"You are adding same feature as string twice: {feature}")
                check_set.add(feature)



from typing import Any, Tuple

from pyarrow import parquet as pyarrow_parquet

from mloda_core.abstract_plugins.components.feature_set import FeatureSet
from mloda_plugins.feature_group.input_data.read_file import ReadFile


class ParquetReader(ReadFile):
    @classmethod
    def suffix(cls) -> Tuple[str, ...]:
        return (
            ".parquet",
            ".PARQUET",
            ".pqt",
            ".PQT",
        )

    @classmethod
    def load_data(cls, data_access: Any, features: FeatureSet) -> Any:
        return pyarrow_parquet.read_table(data_access, columns=list(features.get_all_names()))

    @classmethod
    def get_column_names(cls, file_name: str) -> Any:
        parquet_file = pyarrow_parquet.ParquetFile(file_name)
        return [column.name for column in parquet_file.schema]



from typing import Any, Tuple

from pyarrow import csv as pyarrow_csv

from mloda_core.abstract_plugins.components.feature_set import FeatureSet
from mloda_plugins.feature_group.input_data.read_file import ReadFile


class CsvReader(ReadFile):
    @classmethod
    def suffix(cls) -> Tuple[str, ...]:
        return (
            ".csv",
            ".CSV",
        )

    @classmethod
    def load_data(cls, data_access: Any, features: FeatureSet) -> Any:
        result = pyarrow_csv.read_csv(data_access)
        return result.select(list(features.get_all_names()))

    @classmethod
    def get_column_names(cls, file_name: str) -> Any:
        read_options = pyarrow_csv.ReadOptions(skip_rows_after_names=1)
        table = pyarrow_csv.read_csv(file_name, read_options=read_options)
        return table.schema.names



from typing import Any, Optional, Set, Type
from uuid import UUID

from mloda_core.abstract_plugins.components.feature_name import FeatureName
from mloda_core.abstract_plugins.components.feature import Feature
from mloda_core.abstract_plugins.components.options import Options
from mloda_core.filter.filter_engine import BaseFilterEngine
from mloda_core.filter.single_filter import SingleFilter


class FeatureSet:
    def __init__(self) -> None:
        self.features: Set[Feature] = set()
        self.options: Optional[Options] = None
        # This is just one uuid for easier access
        self.any_uuid: Optional[UUID] = None
        self.filters: Optional[Set[SingleFilter]] = None
        self.name_of_one_feature: Optional[FeatureName] = None
        self.artifact_to_save: Optional[str] = None
        self.artifact_to_load: Optional[str] = None
        self.save_artifact: Optional[Any] = None
        self.filter_engine: Type[BaseFilterEngine] = BaseFilterEngine

    def add_artifact_name(self) -> None:
        if self.options is None:
            raise ValueError("No options set. Call this after adding a feature to ensure Options are initialized.")

        for feature_name in self.get_all_names():
            if feature_name in self.options.data.keys():
                self.artifact_to_load = feature_name
                return

        self.artifact_to_save = self.get_name_of_one_feature().name

    def add(self, feature: Feature) -> None:
        self.features.add(feature)
        self.name_of_one_feature = feature.name
        if self.options is None:
            self.options = feature.options
        if self.any_uuid is None:
            self.any_uuid = feature.uuid

    def remove(self, feature: Feature) -> None:
        self.features.discard(feature)

    def get_all_feature_ids(self) -> Set[UUID]:
        return {feature.uuid for feature in self.features}

    def get_all_names(self) -> Set[str]:
        return {feature.name.name for feature in self.features}

    def __str__(self) -> str:
        return f"{self.features}"

    def get_options(self) -> Options:
        if self.options is None:
            raise ValueError("No options set. Call this after adding a feature to ensure Options are initialized.")
        return self.options

    def get_options_key(self, key: str) -> Any:
        return self.get_options().data.get(key, None)

    def get_initial_requested_features(self) -> Set[FeatureName]:
        return {feature.name for feature in self.features if feature.initial_requested_data}

    def get_name_of_one_feature(self) -> FeatureName:
        if self.name_of_one_feature is None:
            raise ValueError("No feature added yet. Add a feature before calling this method.")
        return self.name_of_one_feature

    def add_filters(self, single_filters: Set[SingleFilter]) -> None:
        if self.filters is not None:
            raise ValueError("Filters already set. This should be called once during setup of the feature set.")

        if not isinstance(single_filters, Set):
            raise ValueError("Filters should be a set.")

        self.filters = single_filters

    def get_artifact(self, config: Options) -> Any:
        return None



import os
from typing import Any

import pyarrow as pa
import sqlite3

from mloda_core.abstract_plugins.components.feature_set import FeatureSet
from mloda_core.abstract_plugins.components.hashable_dict import HashableDict
from mloda_core.abstract_plugins.components.data_types import DataType
from mloda_core.abstract_plugins.components.options import Options
from mloda_plugins.feature_group.input_data.read_db import ReadDB


class SQLITEReader(ReadDB):
    @classmethod
    def db_path(cls) -> str:
        return "sqlite"

    @classmethod
    def connect(cls, credentials: Any) -> Any:
        if isinstance(credentials, HashableDict):
            credential = credentials.data[cls.db_path()]
            return sqlite3.connect(credential)
        else:
            raise ValueError("Credentials must be an HashableDict.")

    @classmethod
    def build_query(cls, features: FeatureSet) -> str:
        query = "select "

        for feature in features.get_all_names():
            query += f"{feature}, "

        query = query[:-2] + " "  # last comma is removed

        query += "from "
        query += f"{cls.get_table(features.options)};"

        if query is None:
            raise ValueError("query cannot be None")
        elif len(query) == 0:
            raise ValueError("query cannot be empty")
        return query

    @classmethod
    def is_valid_credentials(cls, credentials: Any) -> bool:
        """Checks if the given dictionary is a valid credentials object."""

        if isinstance(credentials, HashableDict):
            db_path = credentials.data.get(cls.db_path(), None)
        else:
            db_path = None

        if not isinstance(db_path, str):
            return False

        if not os.path.isfile(db_path):
            raise ValueError(f"Database file {db_path} does not exist, but key is given.")
        return True

    @classmethod
    def load_data(cls, data_access: Any, features: FeatureSet) -> Any:
        query = cls.build_query(features)
        result, column_names = cls.read_db(data_access, query)
        return cls.read_as_pa_data(result, column_names)

    @classmethod
    def read_as_pa_data(cls, result: Any, column_names: Any) -> Any:
        schema = pa.schema([(column_names[i], DataType.infer_arrow_type(result[0][i])) for i in range(len(result[0]))])
        data_dicts = [{column_names[i]: row[i] for i in range(len(row))} for row in result]
        table = pa.Table.from_pylist(data_dicts, schema=schema)
        return table

    @classmethod
    def check_feature_in_data_access(cls, feature_name: str, data_access: Any) -> bool:
        # get tables in the database
        result, _ = cls.read_db(data_access, query="SELECT name FROM sqlite_master WHERE type='table';")
        table_names = [table[0] for table in result]

        # check if the feature_name is in the tables
        for table in table_names:
            result, _ = cls.read_db(data_access, query=f"PRAGMA table_info({table});")
            column_names = [column[1] for column in result]
            if feature_name in column_names:
                cls.set_table_name(data_access, table)
                return True
        return False

    @classmethod
    def set_table_name(cls, data_access: Any, table_name: str) -> None:
        if data_access.data.get("table_name"):
            if data_access.data["table_name"] != table_name:
                raise ValueError(f"Table name is already set to {data_access.data['table_name']} and not {table_name}.")
            return

        data_access.data["table_name"] = table_name

    @classmethod
    def get_table(cls, options: Options | None) -> Any:
        if options is None:
            raise ValueError("Options were not set.")
        return options.get("BaseInputData")[1].data["table_name"]
